package com.platonoff.awt;

import java.awt.Image;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Toolkit;
import java.awt.geom.AffineTransform;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.awt.image.ConvolveOp;
import java.awt.image.CropImageFilter;
import java.awt.image.FilteredImageSource;
import java.awt.image.ImageObserver;
import java.awt.image.Kernel;
import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import javax.swing.ImageIcon;
import com.sun.image.codec.jpeg.JPEGCodec;
import com.sun.image.codec.jpeg.JPEGEncodeParam;
import com.sun.image.codec.jpeg.JPEGImageEncoder;

/**
  JPEG utilities.
  
  Copyright (c) 2002-2004 Dmitry I. Platonoff <dmitry@platonoff.com>
*/

public class JPEGUtil {

  /**
    Loads a given image and creates a thumbnail.
    <P>
    If only one of the dimensions is supplied, the resulting image will
    keep the aspect ratio of the original. If both dimensions are provided
    then an area of the appropriate ratio is cropped from the image and
    then resized to the given dimensions. Also, if the min and max values
    are not equal (which would mean a square thumbnail), an orientation of
    the image is be preserved.
    <P>
    If a sharpen option is used, then the image is resized to a size of
    thumbnail x2, sharpened, and then resized further to the desired size.
    This takes a bit longer, but makes the details more recognizable on
    such a small image.

    @param filename source image file name
    @param maxDimension maximum thumbnail dimension (longer side)
    @param minDimension minimum thumbnail dimension (shorter side)
    @param sharpFactor sharpen coefficient (used as a center value for
      a convolution matrix), 1.0 keeps the original image intact, any
      value greater than 1 causes image sharpening (the more the stronger),
      a value between 0 and 1 causes a blur effect, null means don't do
      anything
    @param cropFactor if not null, means crop the borders of the original
      image before applying the transformation, for example, with a value
      of 0.8 a central area of 80% will make it to the thumbnail (a 10%
      will be cut from each side of the image).
  */
  public static ThumbnailInfo createThumbnail( String filename,
      Integer maxDimension, Integer minDimension,
      Float sharpFactor, Float cropFactor ) throws Exception {
    Image image = new ImageIcon( filename ).getImage();

    // prepare the attributes of the original image
    int srcW = image.getWidth( null );
    int srcH = image.getHeight( null );
    System.err.println( "w=" + srcW + " h=" + srcH );
    float srcRatio = ( ( float ) srcW ) / ( ( float ) srcH );
    if ( srcRatio > 1.0F ) srcRatio = 1.0F / srcRatio;
    boolean widthIsLonger = srcW > srcH;

    // prepare destination attributes
    int destW = 0;
    int destH = 0;
    int cropW = 0;
    int cropH = 0;

    // calculate the destination size and cropping dimensions if needed
    if ( minDimension != null ) {
      if ( maxDimension != null ) {
        if ( maxDimension.intValue() < minDimension.intValue() ) {
          Integer i = maxDimension;
          maxDimension = minDimension;
          minDimension = i;
        }
        float destRatio = ( ( float ) minDimension.intValue() ) /
          ( ( float ) maxDimension.intValue() );
        boolean b = destRatio > srcRatio;

        if ( widthIsLonger ) {
          destW = maxDimension.intValue();
          destH = minDimension.intValue();
          if ( b ) cropW =
            ( int ) ( ( ( float ) ( cropH = srcH ) ) / destRatio );
          else cropH =
            ( int ) ( destRatio * ( float ) ( cropW = srcW ) );
        }
        else {
          destW = minDimension.intValue();
          destH = maxDimension.intValue();
          if ( b ) cropH =
            ( int ) ( ( ( float ) ( cropW = srcW ) ) / destRatio );
          else cropW =
            ( int ) ( destRatio * ( float ) ( cropH = srcH ) );
        }
      }
      else {
        if ( widthIsLonger ) destW = ( int ) ( ( ( float ) ( destH =
          minDimension.intValue() ) ) / srcRatio );
        else destH = ( int ) ( ( ( float ) ( destW =
          minDimension.intValue() ) ) / srcRatio );
      }
    }
    else if ( maxDimension != null ) {
      if ( widthIsLonger ) destH = ( int ) ( srcRatio *
        ( float ) ( destW = maxDimension.intValue() ) );
      else destW = ( int ) ( srcRatio *
        ( float ) ( destH = maxDimension.intValue() ) );
    }
    else throw new IllegalArgumentException(
      "At least one of the dimensions has to be not null." );

    // apply the crop factor
    if ( cropFactor != null ) {
      if ( cropW == 0 ) cropW = srcW;
      if ( cropH == 0 ) cropH = srcH;
      cropW = ( int ) ( cropFactor.floatValue() * ( float ) cropW );
      cropH = ( int ) ( cropFactor.floatValue() * ( float ) cropH );
    }

    class WaitingImageObserver implements ImageObserver {
      boolean mustWait = true;
      public boolean imageUpdate( Image img, int infoflags,
          int x, int y, int width, int height ) {
        return mustWait = ( ( infoflags & ALLBITS ) == 0 );
      }
    };
    WaitingImageObserver wio = new WaitingImageObserver();

    // if there's a need for cropping, crop the image
    if ( cropW > 0 && cropH > 0 ) {
      image = Toolkit.getDefaultToolkit().createImage(
        new FilteredImageSource( image.getSource(),
        new CropImageFilter( ( srcW - cropW ) >> 1,
        ( srcH - cropH ) >> 1, cropW, cropH ) ) );

      image.getWidth( wio );
      for ( int i = 0; i < 50 && wio.mustWait; i++ ) Thread.sleep( 100 );
    }
    else {
      cropW = srcW;
      cropH = srcH;
    }

    if ( sharpFactor != null ) {
      destW <<= 1;
      destH <<= 1;
    }

    // prepare an image buffer
    BufferedImage destImage = new BufferedImage( destW, destH,
      BufferedImage.TYPE_INT_RGB );

    // create an affine transformation
    AffineTransform affineTransform = new AffineTransform();
    affineTransform.scale( ( ( float ) destW ) / ( float ) cropW,
      ( ( float ) destH ) / ( float ) cropH );

    RenderingHints rh = new RenderingHints(
      RenderingHints.KEY_INTERPOLATION,
      RenderingHints.VALUE_INTERPOLATION_BICUBIC );

    // get graphics
    Graphics2D g2d = destImage.createGraphics();
    g2d.setRenderingHints( rh );

    // draw image
    g2d.drawImage( image, affineTransform, wio );
    g2d.dispose();

    // if a sharpen factor is supplied, do the filtering
    if ( sharpFactor != null ) {
      float sharpenValue = ( sharpFactor.floatValue() - 1.0F ) / -4.0F;

      // perform a convolution
      destImage = new ConvolveOp( new Kernel( 3, 3, new float[]
        { 0.0F, sharpenValue, 0.0F, sharpenValue, sharpFactor
        .floatValue(), sharpenValue, 0.0F, sharpenValue, 0.0F } ),
        ConvolveOp.EDGE_NO_OP, null ).filter( destImage,
        new BufferedImage( destW, destH, BufferedImage.TYPE_INT_RGB ) );

      // scale to the final size
      affineTransform = new AffineTransform();
      affineTransform.scale( 0.5F, 0.5F );
      destImage = new AffineTransformOp( affineTransform, rh )
        .filter( destImage, new BufferedImage(
        destW >> 1, destH >> 1, BufferedImage.TYPE_INT_RGB ) );
    }

    return new ThumbnailInfo( destImage, srcW, srcH );
  }


  /**
    Saves an image as JPEG.

    @param file destination file
    @param image image to save
    @param quality image quality from 0.0 to 1.0, where 1 is a lossless
      compression; if null, then the Sun's default will remain untouched
      (the default is 0.5, which, I believe, is too low, especially for
      small images and thumbnails)
  */
  public static void saveImage( File file, BufferedImage image,
      Float quality ) throws IOException {
    FileOutputStream fos = null;
    try {
      JPEGImageEncoder encoder = JPEGCodec.createJPEGEncoder(
        new FileOutputStream( file ) );

      // if a quality is supplied, update the encode parameter
      if ( quality != null ) {
        JPEGEncodeParam jep =
          encoder.getDefaultJPEGEncodeParam( image );
        jep.setQuality( quality.floatValue(), true );
        encoder.setJPEGEncodeParam( jep );
      }

      encoder.encode( image );
    }
    finally { if ( fos != null ) fos.close(); }
  }


   /** Parses the JPEG header to obtain the image size. */
  public static int[] imageSize( File file ) throws IOException {
    FileInputStream fis = null;
    try {
      fis = new FileInputStream( file );
      byte[] b = new byte[4];

      // check for FFD8 in the beginning of file (JPEG signature)
      if ( fis.read( b, 0, 2 ) == 2 &&
          b[0] == ( byte ) 0xFF && b[1] == ( byte ) 0xD8 ) {

        // iterate through data blocks, each JPEG data block starts with
        // a signature FFxx and the following two bytes are the block length
        for ( int n; ( n = fis.read( b, 0, 4 ) ) == 4; ) {
          // check for FFC0 and get two words at offsets 3 and 5
          if ( b[0] == ( byte ) 0xFF && b[1] == ( byte ) 0xC0 ) {
            fis.skip( 1 );
            if ( fis.read( b, 0, 4 ) == 4 ) {
              return new int[]{
                ( ( 0xFF & b[2] ) << 8 ) + ( 0xFF & b[3] ),
                ( ( 0xFF & b[0] ) << 8 ) + ( 0xFF & b[1] ) };
            }
            // otherwise skip to the next block
            else fis.skip( ( ( 0xFF & b[2] ) << 8 ) + ( 0xFF & b[3] ) - 2 );
          }
        }
      }
    }
    finally { if ( fis != null ) fis.close(); }
    return null;
  }


/* test code
  public static void main( String[] args ) throws Exception {
    saveImage( new File( args[1] ), createThumbnail( args[0],
        new Integer( 160 ), new Integer( 120 ), new Float( 2.0F ),
        new Float( 0.8F ) ).thumbnail,
      new Float( 0.9F ) );
    System.exit( 0 );
  }

*/

}
