package com.platonoff.www;

import java.awt.image.BufferedImage;
import java.io.ByteArrayOutputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.util.Iterator;
import java.util.StringTokenizer;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;
import com.platonoff.awt.JPEGUtil;
import com.platonoff.awt.ThumbnailInfo;

/**
  A picture gallery servlet.

  (c) 2001-2004 Dmitry I. Platonoff <dmitry@platonoff.com>
*/
public class PictureServlet extends HttpServlet {

   /** output content type */
  protected final static String CONTENT_TYPE = "x-application/xsl";
   /** default input/output encoding */
  protected final static String ENCODING = "koi8-r";

   // XML elements
  protected final static String PICTURE_SET = "picture-set";
  protected final static String PICTURE_SET_CACHED = "cached";
  protected final static String PICTURE_SET_HAS_PARENT = "has-parent";
  protected final static String PICTURE_SET_THUMBNAIL_LOCATION = "thumbnail-location";
  protected final static String PICTURE_SET_SELECTED = "selected";
  protected final static String PICTURE_SET_EXPAND_ALL = "expand-all";
  protected final static String DIRECTORY = "directory";
  protected final static String DIRECTORY_NAME = "name";
  protected final static String DIRECTORY_TITLE = "title";
  protected final static String PICTURE = "picture";
  protected final static String PICTURE_NAME = "name";
  protected final static String PICTURE_TITLE = "title";
  protected final static String PICTURE_DATE = "date";
  protected final static String PICTURE_PARTIES = "parties";
  protected final static String IMAGE = "image";
  protected final static String IMAGE_WIDTH = "width";
  protected final static String IMAGE_HEIGHT = "height";
  protected final static String THUMBNAIL = "thumbnail";

   // due to the naming convention used, a filename can be broken into parts,
   // containing the date, names and title
  protected final static String partNames[] =
    new String[]{ PICTURE_DATE, PICTURE_PARTIES, PICTURE_TITLE };

   /** config wrapper */
  protected PictureServletConfig config = null;


  public void doGet( HttpServletRequest request,
      HttpServletResponse response ) throws IOException {

    // redirect to a directory explicitly
    if ( ! request.getRequestURI().endsWith( "/" ) ) {
      response.sendRedirect( request.getRequestURI() + "/" );
      return;
    }

    response.setContentType( CONTENT_TYPE + "; charset=" + ENCODING  );

    Element content = createContent( request );

    content.setAttribute( PICTURE_SET_THUMBNAIL_LOCATION,
      config.getThumbnailLocation() );

    // decode path
    String uri = HttpUtil.decode( request.getRequestURI(), ENCODING );

    // extract the current folder name (the part between the last two slashes)
    int pLast = uri.length();
    while ( pLast > 0 && uri.charAt( pLast - 1 ) == '/' ) pLast--;
    int pFirst = ( pLast > 0 ? uri.lastIndexOf( '/', pLast - 1 ) : 0 ) + 1;
    String dirName = getDirName( splitName( pFirst < pLast ?
      uri.substring( pFirst, pLast ) : null, false ) ).replace( '\'', '`' );

    // obtain the selected picture id (for full-size mode)
    String selected = request.getParameter( "i" );

    Element e = createContent( request );
    e.setAttribute( PICTURE_SET_THUMBNAIL_LOCATION,
      config.getThumbnailLocation() );
    if ( selected != null ) e.setAttribute( PICTURE_SET_SELECTED, selected );
    else if ( request.getParameter( "all" ) != null ) {
        e.setAttribute( PICTURE_SET_EXPAND_ALL, String.valueOf( Boolean.TRUE ) );
    }

    e = new Element( "content-box" ).addContent( e );
    e = new Element( "page-content" ).addContent( e );
    e.setAttribute( "title", dirName );
    e = new Element( "page" ).addContent( e );
    String hostName = request.getHeader( "Host" );
    e.setAttribute( "title", dirName + ( hostName == null ? "" :
      " - " + hostName ) );

    new XMLOutputter( "", false, ENCODING )
      .output( new Document( e ), response.getOutputStream() );
  }


  /** 
    A convenience method which can be overridden 
    so that there's no need to call <CODE>super.init(config)</CODE>.
  */
  public void init() {
    config = new PictureServletConfig( getServletContext()
      .getResourceAsStream( PictureServletConfig.CONFIG_PATH ) );
  }


   /** Loads the cached content and updates it if needed. */
  protected Element createContent( HttpServletRequest request ) {
    // decode path
    String uri = HttpUtil.decode( request.getRequestURI(), ENCODING );

    // obtain the real path
    // WARNING: this doesn't work when packed in a .war
    File root = new File( getServletContext().getRealPath( uri ) );
    long lastModified = 0L;

     // load cached content
    Element content = null;
    PictureServletConfig localConfig = config;
    boolean updated = true;
    File contentFile = null;
    try {
      String acceptLanguage = request.getHeader( "Accept-Language" );
      if ( acceptLanguage != null ) {
        StringTokenizer st = new StringTokenizer( acceptLanguage, "," );
        while ( st.hasMoreTokens() ) {
          String s = st.nextToken().trim();
          if ( s.length() > 2 ) s = s.substring( 0, 2 );
          if ( s.equals( config.getDefaultLanguage() ) ) break;
          contentFile = new File( root, config.getContentFileName( s ) );
          if ( contentFile.exists() ) break;
          contentFile = null;
        }
      }
      if ( contentFile == null ) {
        contentFile = new File( root, config.getContentFileName() );
      }
      if ( contentFile.exists() ) {
        content = new SAXBuilder().build( contentFile ).getRootElement();
        try {
          localConfig = ( PictureServletConfig ) config.clone();
          localConfig.init( content );
        }
        catch ( CloneNotSupportedException cloneNotSupportedException ) {
          cloneNotSupportedException.printStackTrace();
        }
      }
    }
    catch ( Exception e ) { e.printStackTrace(); }

    if ( updated = content == null ) {
      // if not found, create a new document
      new Document( content = new Element( PICTURE_SET ) );
      // if not at root, add a "parent folder" directory
      if ( uri.length() > 1 )
        content.setAttribute( PICTURE_SET_HAS_PARENT, "true" );
      content.addContent( "\n" );
    }
    else {
      if ( Boolean.valueOf( content.getAttributeValue( PICTURE_SET_CACHED ) )
        .booleanValue() ) return content;
      lastModified = contentFile.lastModified();
    }

    // get directory listing
    String[] list = root.list();

    // check for existing directories
    for ( Iterator i = content.getChildren( DIRECTORY ).iterator();
        i.hasNext(); ) {
      Element e = ( Element ) i.next();
      int n = findInArray( list, e.getAttributeValue( DIRECTORY_NAME ) );
      if ( n >= 0 && new File( root, list[n] ).isDirectory() ) list[n] = null;
      else {
        i.remove();
        updated = true;
      }
    }

    // check for existing pictures
    for ( Iterator i = content.getChildren( PICTURE ).iterator();
        i.hasNext(); ) {
      Element e = ( Element ) i.next();
      int n = findInArray( list, e.getAttributeValue( PICTURE_NAME ) );
      if ( n >= 0 ) {
        File file = new File( root, list[n] );
        if ( file.isFile() ) {
          if ( file.lastModified() > lastModified ) {
            e.removeChildren( IMAGE );
            e.removeChildren( THUMBNAIL );
            addImage( e, file, localConfig );
            updated = true;
          }
          else {
            File thumbnail = new File( file.getParent() + '/' +
              localConfig.getThumbnailLocation(), file.getName() );
            if ( ! thumbnail.exists() || ( thumbnail.isFile() &&
                thumbnail.lastModified() > lastModified ) ) {
              e.removeChildren( THUMBNAIL );
              addThumbnail( e, file, localConfig );
              updated = true;
            }
          }
          list[n] = null;
          continue;
        }
      }
      i.remove();
      updated = true;
    }

    // iterate through the new elements
    for ( int i = 0; i < list.length; i++ ) if ( list[i] != null ) {
      String name = list[i];
      if ( name.charAt( 0 ) == '.' ) continue;
      if ( name.equalsIgnoreCase( "WEB-INF" ) ) continue;

      if ( ! updated ) updated = true;

      File file = new File( root, name );

      // add new elements
      if ( file.isDirectory() ) addDirectory( content, name );
      else if ( file.isFile() ) addPicture( content, file, localConfig );
    }

     // if the data is updated, save the content file
    if ( updated ) saveContent( contentFile, content );

    return content;
  }


   /** Saves the content. */
  protected void saveContent( File contentFile, Element content ) {
    FileOutputStream fos = null;
    try {
      fos = new FileOutputStream( contentFile );
      new XMLOutputter( "", false, ENCODING )
        .output( content.getDocument(), fos );
    }
    catch ( Exception e ) { e.printStackTrace(); }
    finally {
      if ( fos != null ) try { fos.close(); }
      catch ( IOException ioException ) { ioException.printStackTrace(); }
    }
  }

   /** Adds a directory entry to the list. */
  protected void addDirectory( Element root, String name ) {
    Element e = new Element( DIRECTORY );
    e.setAttribute( DIRECTORY_NAME, name );
    e.setAttribute( DIRECTORY_TITLE, getDirName( splitName( name, false ) ) );

    root.addContent( "  " );
    root.addContent( e );
    root.addContent( "\n" );
  }

   /** Adds a picture entry to the list. */
  protected void addPicture( Element root, File file,
      PictureServletConfig config ) {
    Element e = new Element( PICTURE );
    e.setAttribute( PICTURE_NAME, file.getName() );
    String[] parts = splitName( file.getName(), true );
    for ( int i = parts.length - 1; i >= 0; i-- )
      if ( parts[i] != null ) e.setAttribute( partNames[i], parts[i] );
    e.addContent( "\n  " );

    root.addContent( "  " );
    root.addContent( e );
    root.addContent( "\n" );

    addImage( e, file, config );
  }

   /** Adds an image entry to the picture element. */
  protected void addImage( Element picture, File file,
      PictureServletConfig config ) {
    ThumbnailInfo thumbnailInfo = addThumbnail( picture, file, config );

    Element image = new Element( IMAGE );
    setImageSize( image, file,
      thumbnailInfo == null ? null :
        new int[]{ thumbnailInfo.sourceWidth, thumbnailInfo.sourceHeight } );
    
    picture.addContent( "  " );
    picture.addContent( image );
    picture.addContent( "\n  " );
  }

   /** Adds a thumbnail entry to the picture element. */
  protected ThumbnailInfo addThumbnail( Element picture, File file,
      PictureServletConfig config ) {
    Element thumbnail = new Element( THUMBNAIL );
    File thumbnailFile = new File( file.getParent() + '/' +
      config.getThumbnailLocation(), file.getName() );
    ThumbnailInfo thumbnailInfo = null;

    if ( thumbnailFile.exists() ) {
      setImageSize( thumbnail, thumbnailFile, null );
    }
    else if ( config.isThumbnailGenerationEnabled() ) try {
      File thumbnailDirectory = new File( thumbnailFile.getParent() );
      if ( ! thumbnailDirectory.exists() ) thumbnailDirectory.mkdirs();
      thumbnailInfo = JPEGUtil.createThumbnail(
        file.getPath(),
        config.getMaxThumbnailDimension(),
        config.getMinThumbnailDimension(),
        config.getThumbnailSharpenFactor(),
        config.getThumbnailCropFactor() );
      BufferedImage bi = thumbnailInfo.thumbnail;
      JPEGUtil.saveImage( thumbnailFile, bi,
        config.getThumbnailImageQuality() );
      setImageSize( thumbnail, null,
        new int[]{ bi.getWidth( null ), bi.getHeight( null ) } );
    }
    catch ( Exception e ) { e.printStackTrace(); }

    picture.addContent( "  " );
    picture.addContent( thumbnail );
    picture.addContent( "\n  " );

    return thumbnailInfo;
  }

   /** Obtains an image size and sets the element attributes. */
  protected void setImageSize( Element image, File file, int[] size ) {
    if ( size == null && file != null )
      try { size = JPEGUtil.imageSize( file ); }
      catch ( IOException ioException ) { ioException.printStackTrace(); }

    if ( size != null ) {
      image.setAttribute( IMAGE_WIDTH, String.valueOf( size[0] ) );
      image.setAttribute( IMAGE_HEIGHT, String.valueOf( size[1] ) );
    }
  }

   /** Returns an index of the object in the array or -1. */
  protected static int findInArray( Object[] a, Object o ) {
    for ( int i = a.length - 1; i >= 0; i-- ) {
      if ( a[i] != null && a[i].equals( o ) ) return i;
    }
    return -1;
  }


   /** Constructs a directory name for display from the split parts. */
  protected static String getDirName( String[] dirParts ) {
    StringBuffer dirName = new StringBuffer();
    int n = 0;
    for ( int j = 0; j < dirParts.length; j++ ) if ( dirParts[j] != null ) {
      if ( n++ > 0 ) dirName.append( ": " );
      dirName.append( dirParts[j] );
    }

    return dirName.toString();
  }


  /**
    Splits the filename using a naming convention like this:
    [YYYY[MM[DD]]][nnn][.Name[_Name...]][.Title].jpg
    For example:
    20020420a.Peter_Helga.In_the_garden.jpg
    200204.flowers.jpg
  */
  protected static String[] splitName( String name, boolean isFile ) {
    String[] parts = new String[3];
    if ( name == null ) return parts;

    int p = name.lastIndexOf( '.' );
    parts[2] = ( isFile && p >= 0 ? name.substring( 0, p ) : name )
      .replace( '_', ' ' );

    p = -1;
    if ( Character.isDigit( parts[2].charAt( 0 ) ) && parts[2].length() >= 4 ) {
      StringBuffer date = new StringBuffer( 10 );
      if ( ( p = name.indexOf( '.' ) ) == -1 ) p = parts[2].length();

      if ( p >= 8 ) date.append( name.substring( 6, 8 ) ).append( '.' );
      if ( p >= 6 ) date.append( name.substring( 4, 6 ) ).append( '.' );
      if ( p >= 4 ) date.append( name.substring( 0, 4 ) );
      if ( date.length() > 0 ) parts[0] = date.toString();

      if ( p < parts[2].length() ) parts[2] = parts[2].substring( p + 1 );
      else parts[2] = null;
    }

    if ( parts[2] != null && ( p = parts[2].indexOf( '.' ) ) >= 0 ) {
      parts[1] = parts[2].substring( 0, p ).replace( ' ', '/' );
      parts[2] = parts[2].substring( p + 1 );
    }

    return parts;
  }
}
