package com.platonoff.www;

import java.io.InputStream;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;

/**
  A picture servlet config.
  (c) 2001-2006 Dmitry I. Platonoff <dmitry@platonoff.com>
*/
public class PictureServletConfig implements Cloneable
{
   /** gallery configuration file location */
  final static String CONFIG_PATH =
    "/WEB-INF/picture-gallery-config.xml";

   /** thumbnails section name in the config file */
  protected static final String THUMBNAILS = "thumbnails";
   /** thumbnail location config file parameter */
  protected static final String THUMBNAIL_LOCATION = "location";
   /** thumbnail generation flag config file parameter */
  protected static final String THUMBNAIL_GENERATION_ENABLED = "generate";
   /** max thumbnail dimension config file parameter */
  protected static final String MAX_THUMBNAIL_DIMENSION = "max-dimension";
   /** min thumbnail dimension config file parameter */
  protected static final String MIN_THUMBNAIL_DIMENSION = "min-dimension";
   /** thumbnail sharpen factor config file parameter */
  protected static final String THUMBNAIL_SHARPEN_FACTOR = "sharpen-factor";
   /** thumbnail crop factor config file parameter */
  protected static final String THUMBNAIL_CROP_FACTOR = "crop-factor";
   /** thumbnail JPEG image quality config file parameter */
  protected static final String THUMBNAIL_IMAGE_QUALITY = "jpeg-image-quality";
   /** content file section in the config file */
  protected static final String CONTENT_FILE = "content-file";
   /** content file name config file parameter */
  protected static final String CONTENT_FILE_NAME = "name";
   /** default language section in the config file */
  protected static final String DEFAULT_LANGUAGE = "default-language";
   /** default language config file parameter */
  protected static final String DEFAULT_LANGUAGE_ID = "id";
   /** hotlink image element */
  protected static final String HOTLINK_IMAGE = "hotlink-image";
   /** hotlink image path */
  protected static final String HOTLINK_IMAGE_PATH = "path";
   /** hotlink image content type */
  protected static final String HOTLINK_IMAGE_CONTENT_TYPE = "content-type";
   /** hotlink filter element */
  protected static final String HOTLINK_FILTER = "hotlink-filter";
   /** hotlink filter path */
  protected static final String HOTLINK_FILTER_PATH = "path";
   /** hotlink filter match element */
  protected static final String HOTLINK_FILTER_MATCH = "match";

   /** thumbnail location */
  protected String thumbnailLocation = ".t";
   /** thumbnail generation flag */
  protected boolean thumbnailGenerationEnabled = true;
   /** max thumbnail dimension */
  protected Integer maxThumbnailDimension = null;
   /** min thumbnail dimension */
  protected Integer minThumbnailDimension = null;
   /** thumbnail sharpen factor */
  protected Float thumbnailSharpenFactor = null;
   /** thumbnail crop factor */
  protected Float thumbnailCropFactor = null;
   /** thumbnail JPEG image quality */
  protected Float thumbnailImageQuality = null;
   /** content file name */
  protected String contentFile = ".content.xml";
   /** the name part of the content file name */
  protected String contentFileName = ".content";
   /** the extention part of the content file name */
  protected String contentFileExtention = ".xml";
   /** default language */
  protected String defaultLanguage = "en";

   /** hotlink image path */
  protected String hotlinkImagePath = null;
   /** hotlink image content type */
  protected String hotlinkImageContentType = null;
   /** hotlink path list */
  protected String[] hotlinkPaths = null;
   /** hotlink match map */
  protected Map hotlinkMatches = null;

   /** Returns thumbnail location. */
  public String getThumbnailLocation() { return thumbnailLocation; }
   /** Returns thumbnail generation flag */
  public boolean isThumbnailGenerationEnabled() {
    return thumbnailGenerationEnabled;
  }
   /** Returns max thumbnail dimension. */
  public Integer getMaxThumbnailDimension() { return maxThumbnailDimension; }
   /** Returns min thumbnail dimension. */
  public Integer getMinThumbnailDimension() { return minThumbnailDimension; }
   /** Returns thumbnail sharpen factor. */
  public Float getThumbnailSharpenFactor() { return thumbnailSharpenFactor; }
   /** Returns thumbnail crop factor. */
  public Float getThumbnailCropFactor() { return thumbnailCropFactor; }
   /** Returns thumbnail JPEG image quality. */
  public Float getThumbnailImageQuality() { return thumbnailImageQuality; }
   /** Returns default content file name. */
  public String getContentFileName() { return contentFile; }
   /** Returns content file name for a given languagea. */
  public String getContentFileName( String language ) { 
    return language == null || language.equals( defaultLanguage ) ?
      contentFile : contentFileName + '_' + language + contentFileExtention;
  }
   /** Returns default language. */
  public String getDefaultLanguage() { return defaultLanguage; }
   /** Returns hotlink image path. */
  public String getHotlinkImagePath() { return hotlinkImagePath; }
   /** Returns hotlink image content type. */
  public String getHotlinkImageContentType() { return hotlinkImageContentType; }



   /** Creates a config and initializes it from the specified stream. */
  public PictureServletConfig( InputStream is ) {
    try {
      init( new SAXBuilder().build( is ).getRootElement() );
    }
    catch ( Exception e ) { e.printStackTrace(); }
  }


   /** Initializes the config using the given XML element. */
  public void init( Element root ) {
    loadThumbnails( root );
    loadContentFile( root );
    loadDefaultLanguage( root );
    loadHotlinkFilters( root );
  }


   /** Loads thumbnail configuration. */
  public void loadThumbnails( Element root ) {
    Element e = root.getChild( THUMBNAILS );
    if ( e != null ) {
      String s;
      if ( ( s = e.getAttributeValue( THUMBNAIL_LOCATION ) ) != null ) {
        thumbnailLocation = s;
      }
      if ( ( s = e.getAttributeValue( THUMBNAIL_GENERATION_ENABLED ) ) != null ) {
        thumbnailGenerationEnabled = Boolean.valueOf( s ).booleanValue();
      }

      Integer i;
      if ( ( i = getInteger( e, MAX_THUMBNAIL_DIMENSION ) ) != null ) {
        maxThumbnailDimension = i;
      }
      if ( ( i = getInteger( e, MIN_THUMBNAIL_DIMENSION ) ) != null ) {
        minThumbnailDimension = i;
      }

      Float f;
      if ( ( f = getFloat( e, THUMBNAIL_SHARPEN_FACTOR ) ) != null ) {
        thumbnailSharpenFactor = f;
      }
      if ( ( f = getFloat( e, THUMBNAIL_CROP_FACTOR ) ) != null ) {
        thumbnailCropFactor = f;
      }
      if ( ( f = getFloat( e, THUMBNAIL_IMAGE_QUALITY ) ) != null ) {
        thumbnailImageQuality = f;
      }
    }
  }


   /** Loads the content file name. */
  public void loadContentFile( Element root ) {
    Element e = root.getChild( CONTENT_FILE );
    if ( e != null ) {
      contentFile = e.getAttributeValue( CONTENT_FILE_NAME );
      int n = contentFile.lastIndexOf( '.' );
      if ( n == 0 ) {
        contentFileName = contentFile;
        contentFileExtention = "";
      }
      else {
        contentFileName = contentFile.substring( 0, n );
        contentFileExtention = contentFile.substring( n );
      }
    }
  }
  

   /** Loads the default language setting. */
  public void loadDefaultLanguage( Element root ) {
    Element e = root.getChild( DEFAULT_LANGUAGE );
    if ( e != null ) {
      defaultLanguage = e.getAttributeValue( DEFAULT_LANGUAGE_ID );
    }
  }
  

   /** Loads the hotlink filters. */
  public void loadHotlinkFilters( Element root ) {
    Element e = root.getChild( HOTLINK_IMAGE );
    if ( e != null ) {
      hotlinkImagePath = e.getAttributeValue( HOTLINK_IMAGE_PATH );
      hotlinkImageContentType =
        e.getAttributeValue( HOTLINK_IMAGE_CONTENT_TYPE );
    }

    hotlinkMatches = new HashMap();
    Collection filterElements = root.getChildren( HOTLINK_FILTER );
    hotlinkPaths = new String[filterElements.size()];

    int k = 0;
    for ( Iterator filters = filterElements.iterator(); filters.hasNext(); ) {
      e = ( Element ) filters.next();

      Collection matches = e.getChildren( HOTLINK_FILTER_MATCH );
      String[] matchStrings = new String[matches.size()];

      int j = 0;
      for ( Iterator i = matches.iterator(); i.hasNext(); ) {
        matchStrings[j++] = ( ( Element ) i.next() ).getText();
      }

      String path = e.getAttributeValue( HOTLINK_FILTER_PATH );
      hotlinkMatches.put( path, matchStrings );
      hotlinkPaths[k++] = path;
    }
  }


   /** Checks if any of the existing hotlink filters match the URI. */
  boolean checkForHotlinks( String uri, String refererHost ) {
    for ( int i = hotlinkPaths.length - 1; i >= 0; i-- ) {
      String path = hotlinkPaths[i];
      if ( uri.startsWith( path ) ) {
        String[] matches = ( String[] ) hotlinkMatches.get( path );
        for ( int j = matches.length - 1; j >= 0; j-- ) {
          if ( refererHost.indexOf( matches[j] ) > -1 ) return false;
        }
        return true;
      }
    }
    return false;
  }


   /** Returns an attribute value as an Integer or null. */
  protected static Integer getInteger( Element e, String name ) {
    String s = e.getAttributeValue( name );
    if ( s != null ) try {
      return new Integer( s );
    }
    catch ( NumberFormatException numberFormatException ) { }
    return null;
  }

   /** Returns an attribute value as a Float or null. */
  protected static Float getFloat( Element e, String name ) {
    String s = e.getAttributeValue( name );
    if ( s != null ) try {
      return new Float( s );
    }
    catch ( NumberFormatException numberFormatException ) { }
    return null;
  }

   /** Clones the config object. */
  public Object clone() throws CloneNotSupportedException {
    return super.clone();
  }
}
